//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      This file.                                                       *
//* CLASS    Midwife.java (Midwife)                                           *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     04/05/2014                                                       *
//* REVISION X.0.2                                                            *
//* BYTES    20,558                                                           *
//* TIME     21:44                                                            *
//* PURPOSE  The program Midwife is a Java-coded, CPU/Cluster, load           *
//*          distribution process.  It is premised on the concept that        *
//*          Virtual Agents can be argued to be disembodied "agencies"        *
//*          coupled to some convenient representation, or avatar.  In fact,  *
//*          in a system where avatars serve as vehicular hosts to such       *
//*          "agencies" it is reasonable to assume that the resources of      *
//*          computation, i.e., the processor(s), need not place demands of   *
//*          colocation on either the "agencies" or the host                  *
//*          representations in terms of the Virtual World.  This realization *
//*          suggests that if more than one processor, or CPU, is available   *
//*          in a compute cluster then a load distribution system may help    *
//*          maintain an efficient use of all computation resources during    *
//*          the Virtual World simulation. Midwife provides such a load       *
//*          distribution mechanism.                                          *
//* METHOD   Uses network connectivity and performance metrics observation.   *
//* DETAILS  Programmer notes embedded in source below & CISIS, Rouly, 2014.  *
//* ENVIRON  Developed on a 64-bit Intel Core2Quad, 2.33GHz, with 4GB RAM.    *
//*          The Zotac m-ITX motherboard had board-based Nvidia 9300 graphics *
//*          acceleration. The OS was MS Windows 8 Pro.  Java IDE provided by *
//*          JGrasp version 2.0.0_08, The Java compiler was version 1.7.0_17. *
//*          Verification tests were run in Java JRE build 1.7.0_17-b02 with  *
//*          client VM build 23.7-b01.                                        *
//* COMPILE  javac -g -verbose -deprecation Model.java                        *
//* RUN AS   java -client -Xms1500m -Xmx1500m -Xss1m Model -cp                *
//*          For Windows8 the "Target" line must be: C:\Windows\System32\     *
//*          java.exe -jar "C:\X-Temporary\Personal\Research\Midwife\dist\    *
//*          Midwife.jar /lo=i /hi=j                                          *
//*          The arguments i & j are decimal valued subnet values.  They can  *
//*          be omitted or used.  If used, both should be used.  Their use    *
//*          will speed up network access.                                    *
//*                                                                           *
//*****************************************************************************
//

import java.util.ArrayList;                                   // Java language.
import java.util.Random;                                      //         ditto.
import java.util.Timer;                                       //         ditto.
import java.util.TimerTask;                                   //         ditto.

import network.GetLocalIPA;                                // program routines.
import network.GetSubnetIPAs;                              //            ditto.
import birthing.BirthRoom;                                 //            ditto.
import gui.Console;                                        //            ditto.

//
//*****************************************************************************
// Midwife

class Midwife
{

  //
  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.
  final static boolean USE_FIXED = false;   // start the RNG using a seed type.
  final static long FIXED_SEED = 1681107723981969484L; //fixed data point seed.
  final static long RANDOM_SEED = System.nanoTime (); // fast moving time seed.
  final static long RNG_SEED = USE_FIXED ? FIXED_SEED : RANDOM_SEED; // GLOBAL.
  final static int COMMS_PORT_BASE = 1000; // midwife communications port base.
  final static int PING_PORT = 7;              // port used to find subnetIPAs.
  final static int BACKLOG = 32;    // max queue size for incoming connections.
  final static String AGENT_FILE_NAME = "Agent.exe";  // name of the agent EXE.
  final static int SEC1 = 1000;              // 1 second worth of milliseconds.
  final static int SEC10 = 10000;          // 10 seconds worth of milliseconds.
  final static int SEC20 = 20000;          // 20 seconds worth of milliseconds.
  final static int SEC30 = 30000;          // 30 seconds worth of milliseconds.

  // Variables.
  private static String [] consoleArgs = null;    // console startup arguments.
  private static int activeAgents = 0;    // count of currently started agents.
  private static double t = 0.0D;     // resample interval t is 9.0 - 10.0 sec.
  private static double q = 0.0D;    // resample interval q is 50.0 - 60.0 sec.
  private static double r = 0.0D; // resample interval r is 900.0 - 1000.0 sec.
  private static String myIPA = "";                        // local host's IPA.
  private static ArrayList < String > ipaSubnet = null;     // subnet IPA list.
  private static ArrayList < String > oldIPASubnet = null;  // GUI useful copy.
  private static String osTskMgr = "";      // the name of the OS Task Manager.

  // Objects.
  private static GetLocalIPA localHost = null; // helps get default system IPA.
  private static GetSubnetIPAs cluster = null; // helps get the subnet IPA set.
  private static BirthRoom midwife = null;     // helps instantiate new Agents.
  private static Timer timerGUIUpdate = null; // periodically gets Agents list.
  private static Timer timerAgentList = null; // periodically gets Agents list.
  private static Timer timerIPAList = null;      // periodically gets IPA list.
  private static Timer timerMWLoads = null; // periodically gets midwife loads.
  private static Console console = null;                 // simple command GUI.
  public static Random rng = new Random ( RNG_SEED );//random number generator.

  // --------------------------------------------------------------------------
  /** this method requests the midwife to get the count of agents on this
      CPU and stores the result in the local variable activeAgents. */

  public static void findAgentCounts ( String osTskMgr )
  {

    // remember, osTskMgr helps sort out which MS OS version is being used.
    midwife.countTheAgents ( osTskMgr ); // get agent count active on this CPU.
    activeAgents = midwife.getAgentCount ();  // access the count, store local.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this method queries the cluster for active IPAs and stores the result
      in the local variable ipaSubnet. */

  public static void findClusterIPAs ()
  {

    cluster = new GetSubnetIPAs ( PING_PORT, BACKLOG, myIPA ); // finds subnet.
    cluster.discoverSubnet ( consoleArgs );  // discover active cluster subnet.

    ipaSubnet.clear ();                            // guarantee a clean result.
    ipaSubnet = cluster.getSubnetList ();                // access the results.
    cluster = null;          // return the memory used to the operating system.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this method queries the local host to find its IPA. */

  public static void findLocalIPA ()
  {

    localHost = new GetLocalIPA ();  // a Class to identify the local host IPA.
    localHost.discoverLocalHost ();             // discover the local host IPA.
    myIPA = localHost.getLocalHostStr ();                 // access the result.
    localHost = null;        // return the memory used to the operating system.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this method returns, as a String, the name of the operating system. */

  public static String getOperatingSystemName ()
  {

    final String KEY = "os.name";

    return System.getProperty ( KEY );        // challenge the OS for its name.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this method returns the name of the OS Task Manager executable. */

  public static String getOSTaskMgrExec ( String namedOS )
  {

    final String WIN_XP_CALL = "tasklist.exe"; // assumes using XP Pro version.
    final String WIN_7_CALL = "tasklist.exe";
    final String WIN_8_CALL = "tasklist.exe";
    final String DEFAULT_CALL = "tasklist.exe";    // WARNING: Win types 7 & 8.

    String result = "";
    switch ( namedOS )
    {

      case "Windows XP":
        result = WIN_XP_CALL;
        break;

      case "Windows 7":
        result = WIN_7_CALL;
        break;

      case "Windows 8":
        result = WIN_8_CALL;
        break;

      default:
        System.out.println ( "Error: no OS name recognized in Midwife." );
        result = DEFAULT_CALL;
        break;

    }

    return result;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this reports the subnet surrounding the host IPA. */

  public static void reportSubnetIPAs ()
  {

    int count = ipaSubnet.size ();

    System.out.println ( count + " IPA(s) found in subnet:" );
    for ( int i = 0; i < count; i++ )
    {

      System.out.println ( ipaSubnet.get ( i ) );

    }

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this method initializes the operating environment. */

  public static void setupEnvironment ()
  {

    // Before we begin, which MS Windows OS are we using: XP, 7, or 8?
    String namedOS;                    // the Windows operating system by name.
    namedOS = getOperatingSystemName ();                 // request name of OS.
    Midwife.osTskMgr = getOSTaskMgrExec ( namedOS ); // request OS Tsk Mgr EXE.

    // STEP ONE: Discover the local host IPA.
    findLocalIPA ();                   // query the local host for the its IPA.

    // STEP TWO: Discover who is on the cluster subnet.
    Midwife.ipaSubnet = new ArrayList < String > ();     // establish the list.
    findClusterIPAs ();                   // query the cluster for active IPAs.

    // STEP THREE: Report the cluster subnet (as a list) to the GUI.
    console = new Console ( ipaSubnet );         // this is the user interface.
    console.setEnabled ( true );        // activate the list with cluster IPAs.

    // STEP FOUR: BirthRoom (as midwife) is the central part of the program.
    midwife = new BirthRoom ( COMMS_PORT_BASE, BACKLOG, myIPA, ipaSubnet );
    midwife.setNetListenName ( AGENT_FILE_NAME );      // set the Agent's name.
    findAgentCounts ( Midwife.osTskMgr );      // request midwife count agents.

    // STEP FIVE: Prepare a timer and (re)display the Agent count in GUI.
    t = 9.00D + ( 1.00D * rng.nextDouble () ); // interval 9.00s <= t < 10.00s.
    timerAgentList = new Timer ();                    // let this be the timer.

    // STEP SIX: Prepare a timer that (re)pings the midwives for loads.
    q = 50.0D + ( 10.0D * rng.nextDouble () );  // interval 50.0s <= q < 60.0s.
    timerMWLoads = new Timer ();                      // let this be the timer.

    // STEP SEVEN: Prepare a timer that (re)pings the network for IPA changes.
    r = 900D + ( 1000D * rng.nextDouble () );    // interval 900s <= r < 1000s.
    timerIPAList = new Timer ();                      // let this be the timer.

    // STEP EIGHT: Setup a timer just to maintain the GUI freshness.
    timerGUIUpdate = new Timer ();                    // let this be the timer.
    oldIPASubnet = ipaSubnet;        // match pointers for timerGUI comparison.

    System.gc ();        // clear away any garbage made during the above steps.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this is a timer to periodically reassess the current count of Agents
      under this Midwife's local care.  When the number is found it is force
      displayed on the Console display and it is sent to the BirthRoom midwife
      by synchronized write.
      CURRENTLY: set to fire every 9-10 seconds with initial delay of 10s. */

  public static void timerAgents ( Timer timerAgentList, int delay, double t )
  {

    Timer timer = timerAgentList;
    int period = ( int ) ( t * 1000 );

    timer.scheduleAtFixedRate ( new TimerTask ()
    {

      @Override
      public void run ()
      {

        findAgentCounts ( Midwife.osTskMgr );  // count the agents on this CPU.

      }

    }, delay, period );

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this is a timer to periodically update the GUI mode display.
      CURRENTLY: set to fire every 1 second with initial delay of 0s. */

  public static void timerGUI ( Timer timerGUIUpdate, int delay, double t )
  {

    Timer timer = timerGUIUpdate;
    int period = ( int ) t;

    timer.scheduleAtFixedRate ( new TimerTask ()
    {

      @Override
      public void run ()
      {

        // functionally defined in BirthRoom.countTheAgents().
        final int SYSTEM_FROZEN = Integer.MAX_VALUE;

        boolean freezeXpansion = console.getFreezeXpansion ();  // user choice.
        if ( freezeXpansion )
        {

          console.setAgentCountValue ( SYSTEM_FROZEN ); // force show "Frozen".

        }
        else
        {

          console.setAgentCountValue ( activeAgents );    // display the count.

        }
        midwife.setFreezeXpansion ( freezeXpansion );   // transfer to midwife.

      }
    }, delay, period );

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this is a timer to periodically make the BirthRoom midwife update
      the contents of its "others" Array List.  Once that's done, the midwife
      will have an updated assessment of the current agent IPAs on the subnet
      and of the midwife loadings for all active midwives.
      CURRENTLY: set to fire every 50-60 seconds with initial delay of 30s.*/

  public static void timerMidwifeLoads ( Timer timerMWLoads,
                                         int delay, double q )
  {

    Timer timer = timerMWLoads;
    int period = ( int ) ( q * 1000 );

    timer.scheduleAtFixedRate ( new TimerTask ()
    {

      @Override
      public void run ()
      {

        midwife.updateTheOthersList ();  // cause the updating of the "others".

        // DEBUG - OPTION.
        midwife.reportOthersListContents ();
        // DEBUG - OPTION.

      }

    }, delay, period );

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this is a timer to periodically update the cluster IPA list.  The
      Array List ipaSubnet is force displayed to the console and is made
      available to the BirthRoom midwife by synchronized write.
      CURRENTLY: set to fire every 10 minutes with initial delay of 20s. */

  public static void timerIPAs ( Timer timerIPAList, int delay, double r )
  {

    Timer timer = timerIPAList;
    int period = ( int ) ( r * 1000 );

    timer.scheduleAtFixedRate ( new TimerTask ()
    {

      @Override
      public void run ()
      {

        findClusterIPAs ();      // query the cluster for any new cluster IPAs.

        midwife.setIPASubnet ( ipaSubnet );         // synchro set subnet IPAs.

        if ( !oldIPASubnet.equals ( ipaSubnet ) )
        {

          Midwife.console.updateIPAList ( ipaSubnet );      // update the list.
          oldIPASubnet = midwife.clone ( ipaSubnet );  // make clean duplicate.

        }

      }

    }, delay, period );

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this is the main method. */

  public static void main ( String [] arg )
  {

    consoleArgs = arg;                 // make a copy available for all to use.
    setupEnvironment ();                  // setup all program runtime support.

    timerAgents ( timerAgentList, SEC10, t );      // account for local agents.
    timerIPAs ( timerIPAList, SEC20, r );          // find IPAs in the cluster.
    timerMidwifeLoads ( timerMWLoads, SEC30, q ); // check other midwife loads.
    timerGUI ( timerGUIUpdate, SEC1, SEC1 );           // refresh the user GUI.

    midwife.start ();                                     // start the midwife.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this method does nothing. */

  public void template ()
  {
  } // End method.
  // --------------------------------------------------------------------------

// ----------------------------------------------------------------------------
}
