//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Midwife.java                                                     *
//* CLASS    BirthRoom.java (Birth Room)                                      *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     04/05/2014                                                       *
//* REVISION X.0.2                                                            *
//* BYTES    33,961                                                           *
//* TIME     21:44                                                            *
//* PURPOSE  To provide new agent instantiations, instantiation               *
//*          responsibility auction arbitrage and support, cluster load       *
//*          and loading calculation and prediction.                          *
//* METHOD   Uses network connectivity and performance metrics observation.   *
//* DETAILS  Programmer notes embedded in source below & CISIS, Rouly, 2014.  *
//* ENVIRON  Developed on a 64-bit Intel Core2Quad, 2.33GHz, with 4GB RAM.    *
//*          The Zotac m-ITX motherboard had board-based Nvidia 9300 graphics *
//*          acceleration. The OS was MS Windows 8 Pro.  Java IDE provided by *
//*          JGrasp version 2.0.0_08, The Java compiler was version 1.7.0_17. *
//*          Verification tests were run in Java JRE build 1.7.0_17-b02 with  *
//*          client VM build 23.7-b01.                                        *
//*                                                                           *
//*****************************************************************************
//

package birthing;                                            // home directory.

import java.io.BufferedReader;                                // Java language.
import java.io.File;                                          //         ditto.
import java.io.InputStreamReader;                             //         ditto.
import java.io.IOException;                                   //         ditto.
import java.net.InetAddress;                                  //         ditto.
import java.util.ArrayList;                                   //         ditto.
import java.util.Random;                                      //         ditto.

import server.*;                                              // user packages.

// ----------------------------------------------------------------------------
// BirthRoom

public final class BirthRoom extends Thread
{

  // Constants.
  final static boolean USE_FIXED = false;   // start the RNG using a seed type.
  final static long FIXED_SEED = 1681107723981969484L; //fixed data point seed.
  final static long RANDOM_SEED = System.nanoTime (); // fast moving time seed.
  final static long RNG_SEED = USE_FIXED ? FIXED_SEED : RANDOM_SEED; // GLOBAL.
  final static String AGENTDIR = "Offspring";  // where the Agent code resides.

  // Variables.
  private String agentsName = "";              // this is the name of an Agent.
  private InetAddress selfIPAInetAddr = null;    // will be the local host IPA.
  private String ipa = "";                     // the local host IPA as string.
  private int commsPortBase = 0;  // midwife communications port base, general.
  private int commsPort = 0;     // midwife communications port base, specific.
  private int backlog = 0;          // max queue size for incoming connections.
  private int agentCount = 0;  // the number of Agents active on this host CPU.
  private String localPath = "";     // the canonical path to the Midwife code.
  private File startupFolder = null;  // the folder and path to the Agent code.
  private Boolean freezeXpansion = true;    // control FreezeExpansion display.

  // Objects.
  private ArrayList < String > ipaSubnet = null;        // the subnet IPA list.
  private ArrayList < OtherMidwife > others = null;  // midwives on the subnet.
  public static Random rng = new Random ( RNG_SEED );//random number generator.
  public static Server server = null;     /** this midwife's network interface.

  // --------------------------------------------------------------------------
  /** this private structure is a midwife's internal image of the state of one
      other midwife in the cluster as it existed during the most recent update
      cycle.  The array list called "others" is a container of OtherMidwife.*/

  private class OtherMidwife
  {

    public String ipa = "";                    // the local host IPA as string.
    public int agentCount = 0; // the number of agents active on this host CPU.
    public int bid = 0;        // used in run-off auctions to break a tied bid.

    // a default constructor.
    public OtherMidwife ()
    {
    }

    // a constructor.
    public OtherMidwife ( String othersIPA, int countOfAgents )
    {
      this.ipa = othersIPA;
      this.agentCount = countOfAgents;
      this.bid = 0;
    }

    // returns this Midwife's ipa as a string.
    public String getIPA ()
    {
      return this.ipa;
    }

    // used to set this Midwife's count of active agents.
    public void setCountOfAgents ( int countOfAgents )
    {
      this.agentCount = countOfAgents;
    }

    // used to get this Midwife's count of active agents.
    public int getCountOfAgents ()
    {
      return this.agentCount;
    }

    // used to set this Midwife's tie-breaking bid.
    public void setBid ( int newBid )
    {
      this.bid = newBid;
    }

    // used to get this Midwife's tie-breaking bid.
    public int getBid ()
    {
      return this.bid;
    }

    // used to report to the console facts about this Midwife.
    public void report ()
    {
      System.out.println ( "other's ipa: " + ipa + " agents: " + agentCount );
    }

  } // End class.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this is the constructor. */

  public BirthRoom ( int newPortBase, int newBacklog, String myIPA,
                     ArrayList < String > clusterSubnet )
  {

    this.commsPortBase = newPortBase;  // general port base for communications.

    int offset = getIPAOffset ( myIPA );     // get the 4 from 000.111.222.004.
    this.commsPort = this.commsPortBase + offset;//specific midwife comms port.

    this.backlog = newBacklog;                            // make a local copy.
    this.ipa = myIPA;                                     // make a local copy.
    this.ipaSubnet = clusterSubnet;                       // make a local copy.

    this.agentCount = 0;       // the number of agents active on this host CPU.

    Server server = new Server ();            // instantiate the network tools.
    this.selfIPAInetAddr = server.getIPA ( myIPA );             // get the IPA.
    server = null;                // return the memory to the operating system.

    this.others = new ArrayList < OtherMidwife > (); // midwives on the subnet.

    try
    {

      this.localPath = new File ( "." ).getCanonicalPath ();
      this.startupFolder = new File ( this.localPath + File.separator +
                                      AGENTDIR + File.separator );

    }
    catch ( IOException ioe )
    {

      System.out.println ( "ERROR: Midwife.constructor bad path" );
      System.out.println ( "       " + ioe );

    }

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns a buffered ArrayList that has an added element. */

  public synchronized ArrayList < String > addToQueue
                                ( ArrayList < String > aList, String sParam )
  {

    ArrayList< String> dummy = aList;

    dummy.add ( sParam );
    aList = null;                         // return memory to operating system.

    return dummy;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this takes a String ArrayList, makes a copy, and returns the copy. */

  public synchronized ArrayList < String > clone ( ArrayList < String > list )
  {

    ArrayList < String > clonedList = new ArrayList < String > ();

    for ( int i = 0; i < list.size (); i++ )
    {
      clonedList.add ( list.get ( i ) );
    }

    return clonedList;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this builds a list of all current OS processes, scans the list for
      instances of a particular process named by the variable agentsName,
      counts the instances of that process, and stores that number for
      accessor pickup.  There is provision for selection of OS challenge by
      OS version. */

  public void countTheAgents ( String osTskMgr )
  {

    // used to communicate with user in Console.setAgentCountValue();
    final int SYSTEM_FROZEN = Integer.MAX_VALUE;

    int instancesOfNamedProcess = 0;

    ArrayList < String > processes = getProcessList ( osTskMgr );     // tasks.

    String sample = "";
    int count = processes.size ();
    for ( int i = 0; i < count; i++ )
    {

      sample = processes.get ( i ).trim ();
      if ( sample.equals ( this.agentsName ) )
      {
        instancesOfNamedProcess++;
      }

    }

    if ( this.freezeXpansion )
    {

      this.agentCount = SYSTEM_FROZEN; // set this var to freeze adding Agents.

    }
    else
    {

      this.agentCount = instancesOfNamedProcess;  // set this var for accessor.

    }

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns the number of agents active on this host CPU. */

  public synchronized int getAgentCount ()
  {

    return this.agentCount;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns an ArrayList containing the counts of agents currently
      active on each respective host CPU in the cluster.  To do this, each
      IPA in the cluster subnet is polled via a call to the MsgInterface. */
  public ArrayList < Integer > getClusterLoading
                               ( ArrayList < String > subnet )
  {

    // Sequentially ask each of the midwives on the cluster subnet
    // to report what is the count of agents currently active on
    // their individual host CPU.
    ArrayList < Integer > clusterCounts = new ArrayList < Integer > ();

    int ipaCount = subnet.size ();
    String ipaString = "";            // holds the entire IPA plus whitespaces.
    int offset = 0;                            // holds the last term from IPA.
    int mwLoad = 0;                       // the load a midwife claims to know.
    for ( int i = 0; i < ipaCount; i++ )
    {

      ipaString = subnet.get ( i ).trim ();    // get 1 IPA and trim whtspaces.
      offset = getIPAOffset ( ipaString );   // get the 4 from 000.111.222.004.
      offset = this.commsPortBase + offset; // use the general comms port base.

      mwLoad = server.messaging.getLoadingStatus ( ipaString, offset ); //each.
      if ( mwLoad < 0 )
      {

        // a value less than 0 indicates a NONRESPONSIVE Midwife and one
        // that needs to be prohibited from participating in hosting auctions.
        mwLoad = Integer.MAX_VALUE;

      }
      clusterCounts.add ( new Integer ( mwLoad ) );    // load the output list.

    }
    ipaString = null;                       // return the memory to the system.

    return clusterCounts;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns the offset address in the cluster of the parameter
      ipaString. */

  public int getIPAOffset ( String ipaString )
  {

    Server server = new Server ();                // just for utility purposes.
    int offset = 0;

    // By definition, here in Midwife, the byte in the 32-bit IPA is an offset.
    // Example:  127.000.000.001 will return an "offset" of 1.
    String [] ipaArray = server.split ( ipaString, ".", 4 );         // divide.
    offset = new Integer ( ipaArray [ 3 ] ).intValue ();     // parse on value.

    ipaArray = null;                      // return memory to operating system.
    server = null;                        // return memory to operating system.

    return offset;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this asks the OS for a list of the running processes and returns it
      as an ArrayList of type String.  There is provision for selection of
      OS challenge by OS version. */

  private ArrayList< String > getProcessList ( String osTskMgr )
  {

    ArrayList< String> processList = new ArrayList< String> ();

    Process p = null;
    try
    {

      String line = "";
      p = Runtime.getRuntime ().exec ( osTskMgr + " /fo csv /nh" );

      // mod 02 Jan 2014 - attempt to clear TaskList errors
      wait ( 1500 );  // responding too quickly after call may be bad.
      // mod 02 Jan 2014 - attempt to clear TaskList errors

      BufferedReader input = new BufferedReader (
        new InputStreamReader ( p.getInputStream () ) );

      while ( ( line = input.readLine () ) != null )
      {

        if ( !line.trim ().equals ( "" ) )
        {

          line = line.substring ( 1 );
          processList.add ( line.substring ( 0, line.indexOf ( "\"" ) ) );

        }
      }
      input.close ();

    }
    catch ( IOException ioe )
    {

      System.out.println ( "ERROR: BirthRoom.getProcessList" );
      System.out.println ( "       " + ioe );

    }
    p = null;                             // return memory to operating system.

    return processList;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this helper method returns a simplified String version of the a 32-bit
      IPA.  Given "  010.000.000.012" for example, it will return
      "10.0.0.12." */

  public String ipaSimplifier ( String fullFormat )
  {

    final int MAX_TERMS = 4;                  // the number of terms in an IPA.
    final String DELIMITER = ".";

    String [] stringArray = server.split ( fullFormat.trim (),
                                          DELIMITER, MAX_TERMS );

    Integer intArray [] = new Integer [ stringArray.length ];

    for ( int i = 0; i < stringArray.length; i++ )
    {

      intArray [ i ] = Integer.valueOf ( stringArray [ i ] );

    }

    String simpleFormat = "";
    for ( int i = 0; i < ( intArray.length - 1 ); i++ )
    {

      simpleFormat = simpleFormat + intArray [ i ].toString () + ".";

    }
    simpleFormat = simpleFormat +
                   intArray [ stringArray.length - 1 ].toString ();

    return simpleFormat;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this is an absentee auction process undertaken by the current Midwife
      where all eligible bidders for the responsibility to "birth" a new
      Agency, including the current Midwife, participate by proxy.  The
      assumption is that there is at least one bidder.  Auction with less
      than one bidder is an ERROR.  Midwives having Integer.MAX_VALUE agents
      is an ERROR. */

  public ArrayList < OtherMidwife > proxyAuction
                                    ( ArrayList < OtherMidwife > bidders )
  {

    if ( bidders.size () < 1 )
    {

      System.out.println ( "ERROR: BirthRoom.proxyAuction no bidders" );
      return bidders;

    }
    else if ( bidders.size () == 1 )
    {

      return bidders;

    }
    else
    {

      // start with an empty winners list but these are the candidates to beat.
      ArrayList < OtherMidwife > winners = new ArrayList < OtherMidwife > ();
      int leastLoaded = Integer.MAX_VALUE;   // least likely mW is most loaded.

      for ( OtherMidwife mw : bidders )      // get a Midwife from the cluster.
      {

        if ( mw.getCountOfAgents () < leastLoaded ) // strictly smaller, GREAT!
        {

          winners.clear ();           // start over with an empty winners list.
          winners.add ( mw ); // add new, better-testing candidate to the list.
          leastLoaded = mw.getCountOfAgents ();   //the new least load to beat.

        }
        else if ( mw.getCountOfAgents () == leastLoaded ) // if same smallness.
        {

          winners.add ( mw );   // oops, looks like we may go to a tie-breaker.

        }

      } // obviously, we're ignoring those midwives having larger loads.


      // if there is just one bidder, pass it on through.
      if ( winners.size () == 1 )
      {

        return winners;

      }
      else
      {
        // in the event of a tie, use a proxy auction tie breaker
        winners = prxyActnTieBrkr ( winners );
        for ( OtherMidwife mw : this.others )         // using the GLOBAL list!
        {

          mw.setBid ( 0 );   // after an auction make sure all bids are zeroed.

        }
        return winners;

      }

    }

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this will help proxyAuction break a tie for least loaded status. */

  public ArrayList < OtherMidwife > prxyActnTieBrkr
                                    ( ArrayList < OtherMidwife > bidders )
  {

    ArrayList < OtherMidwife > winner = null;     // ultimately, a list of one.
    int highBid = Integer.MIN_VALUE;              // currently the highest bid.

    do
    {

      winner = null;          // guarantee this is nothing more than a pointer.
      winner = new ArrayList < OtherMidwife > ();

      for ( OtherMidwife bidder : bidders ) // search through the bidders list.
      {

        bidder.setBid ( rng.nextInt () );      // let all bidders randomly bid.

      }

      highBid = Integer.MIN_VALUE;            // reset the current highest bid.
      for ( OtherMidwife bidder : bidders ) // search through the bidders list.
      {

        if ( bidder.getBid () > highBid )     // if bidder has the highest bid.
        {

          highBid = bidder.getBid ();            // wrachet-up the highest bid.

        }

      }

      for ( OtherMidwife bidder : bidders ) // search through the bidders list.
      {

        if ( bidder.getBid () == highBid )    // you're looking for top bidder.
        {

          winner.add ( bidder );    // add all bidders having this bid to list.

        }

      }

    }
    while ( winner.size () != 1 );    // ensure that only one winner emerges.

    return winner;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this reports to the console this Midwife's OthersList contents. */

  public void reportOthersListContents ()
  {

    OtherMidwife anotherMW = null;  // a sample midwife held in the local list.
    String ipaImage = server.ipaStringFormatter ( this.ipa );    // formatting.

    for ( int i = 0; i < this.others.size (); i++ )
    {

      anotherMW = this.others.get ( i );        // a sample list entry to test.

      if ( anotherMW.getIPA ().equals ( this.ipa ) )
      {

        System.out.println ( "self ipa: " + ipaImage +
                             " agents: " + this.agentCount );

      }
      else
      {

        anotherMW.report ();

      }

    }
    anotherMW = null;             // return the memory to the operating system.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this is the BirthRoom's central thread.  It directs its listeners in
      server.messaging to scan for network traffic and requests to start new
      Agents.  New start requests and demands may come from the "Nursery"
      (an external program that starts the initial seed population of agents)
      or from the Agents themselves during runtime as the Agent's breed and
      then wish to have birthed (instantiated) new "infant" Agents.  When
      BirthRoom receives a request to instantiate an Agent it starts a proxy
      auction method to determine which Midwife in the cluster has the least
      CPU loading and thereby, indirectly, the right to instantiate the
      new Agent.  Else, this midwife will birth the new Agent itself. */

  @Override
  public void run ()
  {

    // Every 9-10 seconds this Midwife's agent loading is updated.
    // Every 50-60 seconds all Midwives are polled for their loads.
    // Every 10 minutes the list of Midwive IPAs is updated.

    final String BIRTH_REQUESTED = "instantiationRequest";        // arbitrary.
    final String BIRTH_DEMANDED = "instantiateWithPrejudice";     // arbitrary.
    final String RESERVED = "reserved";                           // arbitrary.

    wait ( 5000 );                 // wait 5000 milliseconds before continuing.

    // top-level network interface object handling inter-midwife messaging.
    server = new Server ( this.ipa, this.commsPort,
                          this.backlog, this.agentCount );

    String currentTasking = RESERVED;   // a reflection of the current taskimg.
    String parameterList = "null";   // the actual OpenSim user parameter list.
    String currentParamList = "";   // a one-deep memory of current parameters.
    do
    {

      server.messaging.setAgentCount ( this.agentCount );   // local load stat.
      server.messaging.main ();  // make a single pass for all message traffic.

      currentTasking = server.messaging.getTasking ();  // what is the tasking.
      parameterList = server.messaging.getParamList ();    // new Agent params.

      // if parameters are NOT eqivalent to the ones we're working on now.
      if ( !currentParamList.equals ( parameterList ) )
      {

        currentParamList = "";       // clear the memory of current parameters.

      }

      if ( currentTasking.equals ( BIRTH_DEMANDED ) &&
           currentParamList.equals ( "" ) )         // prohibits double starts.
      {

        currentParamList = parameterList;   // load memory with current params.
        startLocalAgent ( currentParamList );             // start a new agent.

      }

      if ( currentTasking.equals ( BIRTH_REQUESTED ) &&
           currentParamList.equals ( "" ) ) // prohibits double starts.
      {

        currentParamList = parameterList;   // load memory with current params.

        // after a proxy auction there will be the single winner.
        ArrayList< OtherMidwife> mWWinner = proxyAuction ( this.others );
        String ipaString = mWWinner.get ( 0 ).getIPA ();   // get winner's IPA.
        int port = getIPAOffset ( ipaString ); // get a 4 from 000.111.222.004.
        port = this.commsPortBase + port;   // use the general comms port base.

        // check if self was the winner of the auction.
        String simpleIPA = ipaSimplifier ( ipaString );
        if ( this.ipa.equals ( simpleIPA ) )
        {

          startLocalAgent ( currentParamList );       // choose to start local.

        }
        else
        {

          // choose to start remotely.
          server.messaging.startRemoteAgent ( simpleIPA, port,
                                              currentParamList );
        }

        ipaString = null;                   // return the memory to the system.
        mWWinner = null;                    // return the memory to the system.

      }

    }
    while ( true );

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this synchronized setter sets count of agents active on this host CPU. */

  public synchronized void setAgentCount ( int newCount )
  {

    this.agentCount = newCount;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this sets the value of the user choice to continue expanding Agent
      load. */

  public void setFreezeXpansion ( boolean userChoice )
  {

    this.freezeXpansion = userChoice;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this synchronized setter sets the list of known CPUs in the cluster. */

  public synchronized void setIPASubnet ( ArrayList < String > newList )
  {

    this.ipaSubnet = newList;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this sets the name of the Agents. */

  public void setNetListenName ( String newName )
  {

    this.agentsName = newName;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns a shortened ArrayList.  It takes an ArrayList, drops the
      first element, and moves all other elements forward one.  It can operate
      on lists of many, one, zero or no elements. */

  public synchronized ArrayList < String > shortenTheQueue
                                          ( ArrayList < String > aList )
  {

    if ( aList.isEmpty () )
    {

      aList = null;               // return the memory to the operating system.
      return new ArrayList < String > ();             // just for self-defense.

    }

    if ( aList.size () <= 1 )
    {

      aList = null;               // return the memory to the operating system.
      return new ArrayList < String > ();

    }
    else
    {

      String dummy = aList.remove ( 0 );             // drop the first element.
      dummy = null;               // return the memory to the operating system.

      return aList;

    }

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this method starts an agent object executable locally. */

  private void startLocalAgent ( String startupParticulars )
  {

    // a typical agentsName might be:
    // "Agent.exe"
    // a typical startupParticulars parameter might be:
    // "male G100 password uri:region_X005-Y009&128&128&128 35"

    ProcessBuilder pb = null;              // helps build the OS Agent process.
    Process p = null;                  // pointer to the current Agent process.

    try
    {

      String agentExecutable = this.startupFolder.toString () +
                               File.separator + this.agentsName;

      String[] arg = startupParticulars.split ( " " );
      pb = new ProcessBuilder ( agentExecutable,
                                arg [ 0 ],    // for example: firstName
                                arg [ 1 ],    //              lastName
                                arg [ 2 ],    //              password
                                arg [ 3 ],    //              uri:region&X&Y&Z
                                arg [ 4 ],    //              option 1 (age)
                                arg [ 5 ] );  //              option 2 (other)

      pb.directory ( this.startupFolder );
      p = pb.start ();     // tell the operating system to execute the process.
      arg = null;                 // return the memory to the operating system.
      agentExecutable = null;     // return the memory to the operating system.

    }
    catch ( IOException ioe )
    {

      System.out.println ( "ERROR: Midwife.startLocalAgent" );
      System.out.println ( "       " + ioe );

    }

    pb = null;                    // return the memory to the operating system.
    p = null;                     // return the memory to the operating system.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this builds a structure containing a local image of the state (IPAs and
      momentary agent counts) of the other midwives in the CPU cluster. */

  public void updateTheOthersList ()
  {

    // based on the result of the cluster search in Midwife.timerIPAs, the
    // ipaSubnet ArrayList was produced.  Use it now to create a bufferedList
    // (an IPA-only listing) to help retrieve the this.others list.
    ArrayList < String > bufferedList = clone ( ipaSubnet );
    ArrayList < Integer > countList = getClusterLoading ( bufferedList );

    int countOfCPUs = ipaSubnet.size ();// use the current cluster subnet list.
    String ipa = "";                                        // temporary value.
    int count = 0;                                          // temporary value.

    this.others.clear ();                  // now clear away the current image.
    for ( int i = 0; i < countOfCPUs; i++ )
    {

      ipa = bufferedList.get ( i );        // first (next) IPA in cluster list.
      count = countList.get ( i ).intValue ();     // get the number of agents.

      this.others.add ( new OtherMidwife ( ipa, count ) ); // replace the item.

    }
    // when you leave, the GLOBAL "this.others" is filled with current data.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this causes a wait in milliseconds using system resources. */

  private void wait ( int milliseconds )
  {

    // wait ms before continuing.
    try
    {

      Thread.sleep ( milliseconds );

    }
    catch ( Exception interruptedException )
    {

      System.out.println ( "ERROR: ServerThread.wait" );
      System.out.println ( "       " + interruptedException );

    }

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this method does nothing. */

  public void template ()
  {
  } // End method.
  // --------------------------------------------------------------------------

// ----------------------------------------------------------------------------
}
