//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Midwife.java                                                     *
//* CLASS    GetSubnetIPAs.java (Get Subnet IPAs)                             *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     04/05/2014                                                       *
//* REVISION X.0.2                                                            *
//* BYTES    14,043                                                           *
//* TIME     21:44                                                            *
//* PURPOSE  To discover the active set of IPAs on the subnet near this host. *
//* METHOD   Uses network connectivity and performance metrics observation.   *
//* DETAILS  Programmer notes embedded in source below & CISIS, Rouly, 2014.  *
//* ENVIRON  Developed on a 64-bit Intel Core2Quad, 2.33GHz, with 4GB RAM.    *
//*          The Zotac m-ITX motherboard had board-based Nvidia 9300 graphics *
//*          acceleration. The OS was MS Windows 8 Pro.  Java IDE provided by *
//*          JGrasp version 2.0.0_08, The Java compiler was version 1.7.0_17. *
//*          Verification tests were run in Java JRE build 1.7.0_17-b02 with  *
//*          client VM build 23.7-b01.                                        *
//*                                                                           *
//*****************************************************************************
//

package network;                                             // home directory.

import java.io.BufferedReader;                                // Java language.
import java.io.InputStreamReader;                             //         ditto.
import java.io.IOException;                                   //         ditto.
import java.net.InetAddress;                                  //         ditto.
import java.net.ServerSocket;                                 //         ditto.
import java.util.ArrayList;                                   //         ditto.

import server.*;                                              // user packages.

// ----------------------------------------------------------------------------
// GetSubnetIPAs

public class GetSubnetIPAs
{

  // Constants.
  final boolean LOG_SELF = false;          // add self IPA to the cluster list.
  final int BACKLOG = 32;             // maximum number of pending connections.
  final String OS_PING_EXECUTABLE = "ping";          // an OS specific utility.

  // Variables.
  private int port = 0;                                 // the local host port.
  private int backlog = 0;          // max queue size for incoming connections.
  private String ipa = "";                         // local host IPA as string.

  // Objects.
  private ArrayList < String > ipaSubnet = null;        // an IPA cluster list.
  private Server server = new Server (); // provides network interface service.

  // --------------------------------------------------------------------------
  /** this is the constructor. */

  public GetSubnetIPAs ( int newPort, int newBacklog, String newIPA )
  {

    this.port = newPort;                                  // make a local copy.
    this.backlog = newBacklog;                            // make a local copy.
    this.ipa = newIPA;                                    // make a local copy.

    this.ipaSubnet = new ArrayList < String > ();       // an IPA cluster list.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this scours the 254 subnet network addresses adjoining the base IPA and
  // returns a list of the discovered addresses.  Addresses 1 to 254 are the
  // default span.  Uses PING.EXE to interrogate the space.  Console switches
  // /lo=n or /hi=m inclusive can be used to steer. */

  public void discoverSubnet ( String [] args )
  {

    ServerSocket socket = null;                            // a network socket.
    InetAddress ipaInetAddr = null;                  // testable temporary IPA.

    // this IPA now points to self.
    InetAddress selfIPAInetAddr = server.getIPA ( this.ipa );

    ///////////////////////////////////////////////////////////////////////////
    final String LOCAL_HOST = "127.000.000.001";          // set by convention.
    InetAddress localHost = server.getIPA ( LOCAL_HOST );

    if ( selfIPAInetAddr.equals ( localHost ) )
    {

      this.ipaSubnet.add ( "    " + LOCAL_HOST );   // NOTE: left side padding.
      return;

    }  // this is an early exit for the case of JUST A LOCAL HOST.
    ///////////////////////////////////////////////////////////////////////////

    // create a socket that is now pointing to self.
    socket = server.getServerSocket ( this.port, this.backlog,
                                      selfIPAInetAddr );

    byte [] address = socket.getInetAddress ().getAddress (); // self-centered.

    int loIPA = getLowIPA ( args );
    int hiIPA = getHiIPA ( args );
    for ( int i = loIPA; i <= hiIPA; i++ )
    {

      address[3] = ( byte ) i;  // self-centered IPA that cycles across subnet.
      ipaInetAddr = server.getIPA ( address );    // convert to a testable IPA.

      if ( ping ( ipaInetAddr ) )
      {
        String result [] = ipaInetAddr.toString ().split ( "/" );
        String formattedResult;
        formattedResult = server.ipaStringFormatter (
                          result[ 1 ].toString () );
        this.ipaSubnet.add ( formattedResult );
      }

    }

    server.disconnect ( socket );  // formally disconnect from the socket used.

    ipaInetAddr = null;           // return the memory to the operating system.
    socket = null;                // return the memory to the operating system.
    server = null;                // return the memory to the operating system.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns the values of the discovered IPAs as an Array List. */

  public ArrayList < String > getSubnetList ()
  {

    return this.ipaSubnet;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this attempts to recover and then return the IPA address low range value
  // from the console input argument string. */

  private int getLowIPA ( String [] args )
  {

    final int LOW_LIMIT = 1;
    final String SWITCH_1 = "/lo=";             // 1st console argument switch.

    int lowValue = 0;
    try
    {

      lowValue = Integer.valueOf ( parseStr ( args [ 0 ], SWITCH_1 ) );
      if ( lowValue < LOW_LIMIT )
      {

        lowValue = LOW_LIMIT;

      }

    }
    catch ( ArrayIndexOutOfBoundsException | NumberFormatException e )
    {

      lowValue = LOW_LIMIT;      // a safe value.

    }

    return lowValue;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this attempts to recover and then return the IPA address high range value
  // from the console input argument string. */

  private int getHiIPA ( String [] args )
  {

    final int HIGH_LIMIT = 254;
    final String SWITCH_2 = "/hi=";             // 2nd console argument switch.

    int highValue = 0;
    try
    {

      highValue = Integer.valueOf ( parseStr ( args [ 1 ], SWITCH_2 ) );
      if ( highValue > HIGH_LIMIT )
      {

        highValue = HIGH_LIMIT;

      }

    }
    catch ( ArrayIndexOutOfBoundsException | NumberFormatException e )
    {

      highValue = HIGH_LIMIT;      // a safe value.

    }

    return highValue;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this method parses input passed by the command line to locate and return
  // the value of an argument. */

  public static String parseStr ( String sample, String target )
  {

    int cell = 0;                                 // a pointer into the string.
    String snippet = "";                               // a temporary variable.

    if ( sample.indexOf ( target ) != -1 )       // check if the target exists.
    {

      cell = sample.indexOf ( target );             // where the target begins.
      snippet = sample.substring ( cell );                   // take a snippet.
      cell = cell + target.length ();      // find first cell after the target.
      snippet = snippet.substring ( cell );                  // take a snippet.
      snippet = snippet.trim ();

    }
    else
    {

      System.out.println ( "ERROR: GetSubnetIPAs.parseStr" );

    }

    return snippet;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this uses an Operating System specific "ping" command to check for an
  // active (and DNS port 53 responsive) IPA.  Currently, the assumption is
  // the Operating System is from Microsoft. */

  public boolean ping ( InetAddress ipaInetAddr )
  {

    String pingAdr = ipaInetAddr.toString ();
    pingAdr = pingAdr.substring ( 1, pingAdr.length () );

    String pingString = OS_PING_EXECUTABLE + " " + pingAdr;
    Process p = null;
    try
    {

      p = Runtime.getRuntime ().exec ( pingString );
      BufferedReader inputStream = new BufferedReader (
                                   new InputStreamReader (
                                   p.getInputStream () ) );

      final String SEARCH_TEXT = "bytes=32"; // Microsoft specific return text.

      return pingHelper ( inputStream, SEARCH_TEXT );

    }
    catch ( Exception e )
    {

      System.out.println ( "ERROR: GetSubnetIPAs.ping" );

    }
    p = null;                             // return memory to operating system.

    System.out.println ( "ERROR: ping failed." );

    return false;  // major problem, likely OS version troubles.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this processes the data in the BufferedReader input passed to it
  // and returns a Boolean true if the String searchText can be found. */

  private boolean pingHelper ( BufferedReader input, String searchText )
  {

    // assume the text is not here.
    try
    {

      CharSequence cSeqPositive = searchText.subSequence ( 0, searchText.
        length () );
      CharSequence cSeqNegative = "Destination host unreachable";
      String line = "";

      while ( ( line = input.readLine () ) != null )
      {
        if ( line.contains ( cSeqPositive ) )
        {
          return true; // oh, it is here.
        }
        if ( line.contains ( cSeqNegative ) )
        {
          return false; // oh, it is not here.
        }
      }
      input.close ();

    }
    catch ( IOException ioe )
    {

      System.out.println ( "ERROR: GetSubnetIPAs.pingHelper" );
      System.out.println ( "       " + ioe );

    }

    return false; // nope, not here.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this method does nothing. */

  public void template ()
  {
  } // End method.
  // --------------------------------------------------------------------------

// ----------------------------------------------------------------------------
}
