//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Server.java (Server)                                             *
//* CLASS    MsgInterface.java (Message Interface)                            *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     04/05/2014                                                       *
//* REVISION X.0.2                                                            *
//* BYTES    29,243                                                           *
//* TIME     21:44                                                            *
//* PURPOSE  To transmit and receive network communications that support      *
//*          inter-Midwife, Nursery-to-Midwife, and Agency-to-Midwife         *
//*          network traffic.                                                 *
//* METHOD   Listens to network communication and supports inter-Midwife,     *
//*          Nursery-to-Midwife, and Agency-to-Midwife network comms.         *
//* DETAILS  Programmer notes embedded in source below & CISIS, Rouly, 2014.  *
//* ENVIRON  Developed on a 64-bit Intel Core2Quad, 2.33GHz, with 4GB RAM.    *
//*          The Zotac m-ITX motherboard had board-based Nvidia 9300 graphics *
//*          acceleration. The OS was MS Windows 8 Pro.  Java IDE provided by *
//*          JGrasp version 2.0.0_08, The Java compiler was version 1.7.0_17. *
//*          Verification tests were run in Java JRE build 1.7.0_17-b02 with  *
//*          client VM build 23.7-b01.                                        *
//*                                                                           *
//*****************************************************************************
//

package server;                                              // home directory.

import java.io.IOException;                                   // Java language.
import java.io.InputStream;                                   //         ditto.
import java.io.OutputStream;                                  //         ditto.
import java.net.InetAddress;                                  //         ditto.
import java.net.ServerSocket;                                 //         ditto.
import java.net.Socket;                                       //         ditto.

// ----------------------------------------------------------------------------
public final class MsgInterface
{

  // if a tested IPA (a midwife) fails a network transaction its health
  // (mwHealth) is set to NONRESPONSIVE.  The variables (GLOBAL) mwHealth,
  // response, and (CONSTANT) NONRESPONSIVE are responsible for data flow.

  // Constants.
  final String RESERVED = "reserved";                             // arbitrary.
  final String STATUS_REQUESTED = "reportActiveCount";            // arbitrary.
  final String STATUS_REPORT = "activeAgentCount=";               // arbitrary.
  final String BIRTH_REQUESTED = "instantiationRequest";          // arbitrary.
  final String BIRTH_DEMANDED = "instantiateWithPrejudice";       // arbitrary.
  final String ACKNOWLEGEMENT = "understoodProceed";              // arbitrary.
  final String SAY_PARAMETER_LIST = "parameter list#";            // arbitrary.
  final String REQUEST_REFUSED = "request_refused";               // arbitrary.
  final int NONRESPONSIVE = Integer.MAX_VALUE;   // used for nonresponsive IPA.

  // Variables.
  private String ipa = "";                     // the local host IPA as string.
  private int mwHealth = -1; // reflects on a midwife's health, assume failure.
  private int agentCount = 0;                      // number of current agents.
  private String readBuffer = "";    // (GLOBAL) current inter-Midwife message.
  private String loadStatus = "";    // (GLOBAL) current inter-Midwife message.
  private String startAgent = "";    // (GLOBAL) current inter-Midwife message.
  private String currentTask = "";            // the Midwife's current tasking.
  private String paramList = "";         // the parameters used to make agents.

  // Objects.
  private static ServerSocket serverSocket = null;         // a default socket.
  private static Socket socket = null;                    // a specific socket.

  // --------------------------------------------------------------------------
  /** this is the constructor. */

  public MsgInterface ( ServerSocket newServerSocket, int currentCount,
                        String hostIPA )
  {

    serverSocket = newServerSocket;         // this interface's default socket.
    this.agentCount = currentCount;                // number of current agents.
    this.ipa = hostIPA;                        // the local host IPA as string.

    this.readBuffer = "";                                        // initialize.
    this.loadStatus = "";                                        // initialize.
    this.startAgent = "";                                        // initialize.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  // WARNING: long method.
  /** this is a branching, CONSTANT matching, Boolean evaluator.  Each
      conditional branch corresponds to one valid inter-Midwife messaging
      transaction pair - and explicit conversation.  The private currentTask
      and paramList variables are assigned here.  They are accessor
      available. */

  private synchronized void evaluateMidwifeMessage ( InputStream inStream,
                                                     OutputStream outStream )
  {

    boolean dummy = false;             // a useless temporary holding variable.
    this.currentTask = RESERVED;                                // clean reset.
    this.paramList = "null";                                    // clean reset.

    //-----------------  MESSAGE (m0) RESERVED  --------------
    if ( this.readBuffer.equals ( RESERVED ) )
    {

      this.currentTask = RESERVED;           // GLOBAL name of current tasking.
      send ( RESERVED, outStream );            // a reserved value, no meaning.
      this.paramList = "null";                               // agent reqments.

    }
    //-----------------  MESSAGE (m1) STATUS REQUESTED  -----
    else if ( this.readBuffer.equals ( STATUS_REQUESTED ) )
    {

      this.currentTask = STATUS_REQUESTED;   // GLOBAL name of current tasking.
      Integer count = new Integer ( this.agentCount );
      String answer = STATUS_REPORT + count.toString ();
      send ( answer, outStream );                      // quickly, send it out.
      this.paramList = "null";                               // agent reqments.
      count = null;                         // return the memory to the system.

    }
    //-----------------  MESSAGE (m2) BIRTH DEMANDED  -------
    else if ( this.readBuffer.equals ( BIRTH_DEMANDED ) )
    {

      this.currentTask = BIRTH_DEMANDED;     // GLOBAL name of current tasking.
      send ( ACKNOWLEGEMENT, outStream );          // quickly, send a response.
      dummy = receive ( inStream, "evaluateMidwifeMessage" );   // read buffer.
      this.paramList = getArguments ( this.readBuffer );     // agent reqments.

    }
    //-----------------  MESSAGE (m3) BIRTH REQUESTED  -------
    else if ( this.readBuffer.equals ( BIRTH_REQUESTED ) )
    {

      this.currentTask = BIRTH_REQUESTED;    // GLOBAL name of current tasking.
      send ( ACKNOWLEGEMENT, outStream );          // quickly, send a response.
      dummy = receive ( inStream, "evaluateMidwifeMessage" );   // read buffer.
      this.paramList = getArguments ( this.readBuffer );     // agent reqments.

    }
    //-----------------  MESSAGE (m-) ------------------------
    else
    {

      // currently, an unsupported option.
      this.currentTask = RESERVED;           // GLOBAL name of current tasking.
      send ( RESERVED, outStream );            // a reserved value, no meaning.
      this.paramList = "null";                               // agent reqments.
      System.out.println ( "ERROR: MsgInterface.evaluateMidwifeMessage" );
      System.out.println ( "       readBuffer contained: " + this.readBuffer );

    }
    //--------------------------------------------------------

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this flushes data in network buffers associated with the IO streams. */

  private void flushBuffers ( OutputStream outStream,
                              InputStream inStream )
  {

    try
    {

      outStream.flush ();
      inStream.close ();
      outStream.close ();

    }
    catch ( Exception ioException )
    {

      System.out.println ( "ERROR: MsgInterface.flushBuffers" );
      System.out.println ( "       " + ioException );

    }

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns the string arguments embedded in the buffered parameter
      describing the requested Agency (instantiation). */

  public String getArguments ( String bufferContents )
  {

    // for reference only: SAY_PARAMETER_LIST = "parameter list#"
    // FOR EXAMPLE:
    // "parameter list#male G100 password uri:region_X005-Y009&128&128&128 35"

    String result = "";                // the actual parameter list to be used.

    String array [] = bufferContents.split ( "#" );
    result = array [ 1 ];
    array = null;                           // return the memory to the system.

    return result;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns the last message received in inter-Midwife communications.*/

  public String getTasking ()
  {

    return this.currentTask;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns the parameter list of new agent instantiation particulars.*/

  public String getParamList ()
  {

    return this.paramList;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  // WARNING: long method.
  /** this is the load status requester.  It directs the named Midwife to
      reveal its current agent load.  It returns a value of 0 (or greater)
      that is equivalent to the respective agent load report (if succesful).
      It returns -1 if it fails to get data from a midwifeIPA. */

  public int getLoadingStatus ( String midwifeIPA, int port )
  {

    // check the main method in-operations flag via synchronized accessor.
    Socket s = null;      // a special socket used only to request load status.
    InputStream iStream = null;                // pointer to the stream object.
    OutputStream oStream = null;               // pointer to the stream object.

    int response = -1;    // assume testable midwife IPA will be nonresponsive.
    this.mwHealth = -1;     // initially -1 reflects midwife health neutrality.

    if ( testServerListening ( midwifeIPA, port ) )
    {

      try
      {

        //System.out.println ( "Attempting to connect with midwife at" );
        //System.out.println ( "IPA is: " + midwifeIPA + " port: " + port );
        s = new Socket ( midwifeIPA, port );

      }
      catch ( IOException ioe )
      {

        this.mwHealth = NONRESPONSIVE;    // GLOBAL used for nonresponsive IPA.
        System.out.println ( "ERROR: MsgInterface.getLoadingStatus 1" );
        System.out.println ( "       error was: " + ioe.toString () );

      }
      catch ( NullPointerException npe )
      {

        this.mwHealth = NONRESPONSIVE;    // GLOBAL used for nonresponsive IPA.
        System.out.println ( "ERROR: MsgInterface.getLoadingStatus 2" );
        System.out.println ( "       error was: " + npe.toString () );

      }

      try
      {

        iStream = s.getInputStream ();
        oStream = s.getOutputStream ();

      }
      catch ( IOException ioe )
      {

        this.mwHealth = NONRESPONSIVE;    // GLOBAL used for nonresponsive IPA.
        System.out.println ( "ERROR: MsgInterface.getLoadingStatus  3" );
        System.out.println ( "       error was: " + ioe.toString () );

      }

      // send a request for load status.
      // for reference only: STATUS_REQUESTED = "reportActiveCount"
      boolean transmissionOK = send ( STATUS_REQUESTED, oStream );

      // if the transmission went out OK wait to receive the answer.
      if ( transmissionOK )
      {

        transmissionOK = receive ( iStream, "getLoadingStatus" );   // go read!

      }
      else
      {

        System.out.println ( "ERROR: MsgInterface.getLoadingStatus  4" );

      }

      // if you're here, the request for loading was answered, try to decypher
      // the message you received - it's in the GLOBAL loadStatus buffer.
      if ( transmissionOK && ( this.loadStatus != null ) )
      {

        // FOR EXAMPLE: "activeAgentCount=1234"
        String array [] = this.loadStatus.split ( "=" );
        response = new Integer ( array [ 1 ] ).intValue ();
        array = null;                      // return the memory to the system.

      }
      else
      {

        this.mwHealth = NONRESPONSIVE;    // GLOBAL used for nonresponsive IPA.
        System.out.println ( "ERROR: load request from " + midwifeIPA );
        System.out.println ( " unavailable at port " + port );

      }

      flushBuffers ( oStream, iStream );
      iStream = null;                      // return the memory to the system.
      oStream = null;                      // return the memory to the system.

    }

    // evaluate load status.  The health of the midwife is an indirect sign.
    // Response "healthy" range is 0 to "n" but less than int.NONRESPONSIVE.
    if ( response > this.mwHealth )          // mwHealth always starts at a -1.
    {                            // else, for a failure, mwHealth is MAX'd out.

      return response;    // this was a healthy midwife had zero or more loads.

    }
    else
    {

      response = -1;  // having been tested for failure and failed report same.

    }

    return response;                 // response will be re-intialized to a -1.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this is the MsgInterface main execution thread.  It begins by accepting
      the serverSocket connection passed in from above.  Then it ties itself
      to Input and Output Streams.  It directs a network blocking receiver to
      wait for a new command message the receipt and execution of which is the
      whole purpose of the thread.  If there is data on the network, it will
      have been placed in a buffer variable by the time main regains control.
      This then directs the evaluation of the data. */

  public void main ()
  {

    InputStream inStream = null;      // byte stream to character stream.
    OutputStream outStream = null;    // character stream to byte stream.

    try
    {

      // WARNING:  This method blocks until a connection is made.
      socket = serverSocket.accept ();
      // WARNING:  This method blocks until a connection is made.

    }
    catch ( IOException ioe )
    {

      System.out.println ( "ERROR: MsgInterface.main 1" );
      System.out.println ( "       error was: " + ioe.toString () );

    }

    try
    {

      // oddly enough, outstream MUST preceed instream!!!!
      outStream = socket.getOutputStream ();
      inStream = socket.getInputStream ();

    }
    catch ( IOException ioException )
    {

      System.out.println ( "ERROR: MsgInterface.main 2" );

    }

    if ( receive ( inStream, "main" ) )            // blocking read of network.
    {

      evaluateMidwifeMessage ( inStream, outStream );     // conditional evals.

    }
    else
    {

      //System.out.println ( "NOTICE: MsgInterface.main network buffer was " );
      //System.out.println ( "        empty.  No message evaluation done." );
      this.currentTask = RESERVED;                              // clean reset.
      this.paramList = "null";                                  // clean reset.

    }

    flushBuffers ( outStream, inStream );
    inStream = null;                        // return the memory to the system.
    outStream = null;                       // return the memory to the system.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this method blocks while receiving data from the network interface,
      converts the received data (when/if it arrives) into a String, stores
      what was received in the GLOBAL readBuffer and then returns True if
      the activity was successful.  Else, it returns false. */

  private boolean receive ( InputStream inStream, String user )
  {

    final int EOS = -1;                                       // End Of Stream.
    final int BUFFER_SIZE = 1024;                      // buffer size in bytes.

    byte [] netBuffer = new byte [ BUFFER_SIZE ];   // the network read buffer.
    boolean status = false;                   // assume the activity will fail.

    try
    {

      // WARNING:  This blocks until something is received.
      while ( !status && ( inStream.read ( netBuffer ) != EOS ) )
      {

        // converts byte[] to String and trims away leading and trailing
        // white spaces all in synchronized, accessor readable locations.
        setReadBuffers ( netBuffer, user );
        status = true;

      }

    }
    catch ( IOException ioe )
    {

      status = false;
      this.readBuffer = "";      // ERROR condition on a read from the network.
      this.mwHealth = NONRESPONSIVE;      // GLOBAL used for nonresponsive IPA.
      System.out.println ( "ERROR: Exception in MsgInterface.receive" );

    }

    netBuffer = null;             // return the memory to the operating system.
    return status;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this converts a String message (botschaft) into a byte array
      and then transmits that array over the network.  It reports status:
      True if the activity was successful, else, false. */

  private boolean send ( String botshaft, OutputStream outStream )
  {

    boolean status = false;                   // assume the activity will fail.

    // attempt to send the object.
    try
    {

      outStream.write ( botshaft.getBytes () );  // send a convertd byte array.
      outStream.flush ();                                          // clean up.
      status = true;                                // activity was successful.

    }
    catch ( Exception ioException )
    {

      System.out.println ( "ERROR: MsgInterface.send" );

    }

    return status;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this sets the (local, reportable, and current) agent count. */

  public void setAgentCount ( int count )
  {

    this.agentCount = count;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this sets the data into the network read after recieve GLOBAL buffers. */

  private synchronized void setReadBuffers ( byte[] data, String user )
  {

    // converts byte[] to String, trims away leading and trailing
    // white spaces, and put it all into GLOBAL readable locations.
    String dummy = ( new String ( data ) ).trim ();

    if ( user.equals ( "main" ) ||
         user.equals ( "evaluateMidwifeMessage" ) )
    {
      this.readBuffer = dummy;
    }

    if ( user.equals ( "getLoadingStatus" ) )
    {
      this.loadStatus = dummy;
    }

    if ( user.equals ( "startRemoteAgent" ) )
    {
      this.startAgent = dummy;
    }

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  // WARNING: long method.
  /** this sends a message to a remote Midwife that DEMANDS a new agent
      start. */

  public void startRemoteAgent ( String ipa, int port,
                                 String parameterList )
  {

    Socket s = null;      // a special socket used only to start remote Agents.
    InputStream iStream = null;                // pointer to the stream object.
    OutputStream oStream = null;               // pointer to the stream object.

    try
    {

      //System.out.println ( "Attempting to connect with midwife at" );
      //System.out.println ( "IPA is: " + ipa + " port: " + port );
      s = new Socket ( ipa, port );

    }
    catch ( IOException ioe )
    {

      System.out.println ( "ERROR: MsgInterface.startRemoteAgent 1" );
      System.out.println ( "       error was: " + ioe.toString () );

    }
    catch ( NullPointerException npe )
    {

      System.out.println ( "ERROR: MsgInterface.startRemoteAgent 2" );
      System.out.println ( "       error was: " + npe.toString () );

    }

    try
    {

      iStream = s.getInputStream ();
      oStream = s.getOutputStream ();

    }
    catch ( IOException ioe )
    {

      System.out.println ( "ERROR: MsgInterface.startRemoteAgent 3" );
      System.out.println ( "       error was: " + ioe.toString () );

    }

    // send a demand for a new agent instantiation.
    // for reference only: BIRTH_DEMANDED = "instantiateWithPrejudice"
    boolean transmissionOK = send ( BIRTH_DEMANDED, oStream );

    // if the transmission is not successful print ERROR msg.
    if ( transmissionOK )
    {

      transmissionOK = receive ( iStream, "startRemoteAgent" );        // read.

      // for reference only: ACKNOWLEGEMENT = "understoodProceed"
      if ( transmissionOK && this.startAgent.equals ( ACKNOWLEGEMENT ) )
      {

        // for reference only: SAY_PARAMETER_LIST = "parameter list#"
        String botshaft = SAY_PARAMETER_LIST + parameterList;

        // transmit the data about the new agent to the remote midwife.
        transmissionOK = send ( botshaft, oStream );

      }
      else
      {

        System.out.println ( "ERROR: startRemoteAgent failed 1" );

      }

    }
    else
    {

      System.out.println ( "ERROR: startRemoteAgent failed 2" );

    }

    flushBuffers ( oStream, iStream );
    iStream = null;                       // return memory to operating system.
    oStream = null;                       // return memory to operating system.

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this is test for a listening server on a specific host IPA and port. */

  public boolean testServerListening ( String host, int port )
  {

    InetAddress argA = null;                // a pointer to a convertible form.
    InetAddress argB = null;                // a pointer to a convertible form.
    try
    {

      argA = InetAddress.getByName ( this.ipa );  // convert from string to ...
      argB = InetAddress.getByName ( host );      // convert from string to ...

    }
    catch ( Exception ioException )
    {

      System.out.println ( "ERROR: msgInterface.testServerListening 1" );
      System.out.println ( "       " + ioException );

    }

    if ( argA.equals ( argB ) )                        // is this a self-check.
    {
      return true;         // if the two IPAs are equivalent take a quick exit.
    }
    else
    {

      Socket s = null;                     // let's get ready to check the IPA.

      try
      {
        //System.out.println ( "Attempting to verify server listening at" );
        //System.out.println ( "host is: " + host + " port: " + port );
        s = new Socket ( host, port );                           // OK, try it.
        s.close ();                       // if there was no failure, close it.
        return true;  // super, the IPA was valid (or the firewall porous, ha!)

      }
      catch ( Exception e )
      {

        // DEBUG ONLY - this is the expected error caused by a FAILED IPA test!
        //System.out.println ( "ERROR: msgInterface.testServerListening 2" );
        //System.out.println ( "       " + e );
        return false;                       // report the failure as a Boolean.

      }
      finally                        // what follows is just a little clean up.
      {

        if ( s != null ) // in case of a socket that is undefined but possible.
        {

          try
          {

            s.close ();                 // close it for safety and completeness.

          }
          catch ( Exception e )
          {

            System.out.println ( "ERROR: msgInterface.testServerListening 3" );
            System.out.println ( "       " + e );

          }

        }

      }

    }

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this causes a wait in milliseconds using system resources. */

  private void wait ( int milliseconds )
  {

    // wait ms before continuing.
    try
    {

      Thread.sleep ( milliseconds );

    }
    catch ( Exception interruptedException )
    {

      System.out.println ( "ERROR: ServerThread.wait" );
      System.out.println ( "       " + interruptedException );

    }

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this method does nothing. */

  public void template ()
  {
  } // End method.
  // --------------------------------------------------------------------------

// ----------------------------------------------------------------------------
}
