//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Midwife.java                                                     *
//* CLASS    Server.java (Server)                                             *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     04/05/2014                                                       *
//* REVISION X.0.2                                                            *
//* BYTES    16,575                                                           *
//* TIME     21:44                                                            *
//* PURPOSE  To provide Midwife with several simple networking utilities.     *
//*          For example, the class has methods capable of quickly and        *
//*          safely instantiating sockets (ServerSockets and simple Sockets), *
//*          testing them, and evaluating them.  It provides execution        *
//*          exception handling away from the main program execution path     *
//*          and thus "de-clutters" the readability of the more central       *
//*          elements of the Midwife source code and its components.          *
//* METHOD   Uses network connectivity and performance metrics observation.   *
//* DETAILS  Programmer notes embedded in source below & CISIS, Rouly, 2014.  *
//* ENVIRON  Developed on a 64-bit Intel Core2Quad, 2.33GHz, with 4GB RAM.    *
//*          The Zotac m-ITX motherboard had board-based Nvidia 9300 graphics *
//*          acceleration. The OS was MS Windows 8 Pro.  Java IDE provided by *
//*          JGrasp version 2.0.0_08, The Java compiler was version 1.7.0_17. *
//*          Verification tests were run in Java JRE build 1.7.0_17-b02 with  *
//*          client VM build 23.7-b01.                                        *
//*                                                                           *
//*****************************************************************************
//

package server;                                              // home directory.

import java.io.IOException;                                   // Java language.
import java.net.InetAddress;                                  //         ditto.
import java.net.ServerSocket;                                 //         ditto.
import java.net.Socket;                                       //         ditto.
import java.net.UnknownHostException;                         //         ditto.
import java.util.ArrayList;                                   //         ditto.
import javax.net.ServerSocketFactory;                         //         ditto.

// ----------------------------------------------------------------------------
// Server

public class Server extends Thread
{

  // Constants.

  // Variables.
  private int agentCount = 0;                      // number of current agents.

  // Objects.
  private static ServerSocketFactory socketFactory =
                                     ServerSocketFactory.getDefault ();
  private static ServerSocket surrogateSocket = null; // a stand-in for socket.
  public MsgInterface messaging = null;      /** this is the network interface.

  // --------------------------------------------------------------------------
  /** this is a default constructor. */

  public Server ()
  {
  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this is the server thread constructor. */

  public Server ( String hostIPA, int commsPort, int backlog,
                  int currentCount )
  {

    this.agentCount = currentCount;                // number of current agents.

    // create an instance of a server socket using constraints.
    // the hostIPA is the network adapter IPA address, then
    // for example: if hostIPA were 127.0.0.1, then ...
    // for example: this.commsPort would be 1001, and ...
    // for example: bindAddress would be /127.0.0.1

    try
    {

      InetAddress bindAddress = InetAddress.getByName ( hostIPA );
      surrogateSocket = socketFactory.createServerSocket ( commsPort,
                                                           backlog,
                                                           bindAddress );

    }
    catch ( Exception ioException )
    {

      System.out.println ( "ERROR: Server.Server" );
      System.out.println ( "       " + ioException );

    }

    messaging = new MsgInterface ( surrogateSocket, this.agentCount, hostIPA );

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this attempts to release the server thread socket connectedSocket. */

  private void disconnect ()
  {

    try
    {

      surrogateSocket.close ();
      surrogateSocket = null;         // return memory to the operating system.

    }
    catch ( IOException ioe )
    {

      System.out.println ( "ERROR: Server.disconnect" );
      System.out.println ( "       " + ioe );

    }

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this attempts to release the socket connectedSocket. */

  public void disconnect ( Socket socket )
  {

    try
    {

      socket.close ();
      socket = null;                  // return memory to the operating system.

    }
    catch ( IOException ioe )
    {

      System.out.println ( "ERROR: Server.disconnect (Socket)" );
      System.out.println ( "       " + ioe );

    }

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this attempts to release the server socket connectedSocket. */

  public void disconnect ( ServerSocket socket )
  {

    try
    {

      socket.close ();
      socket = null;                  // return memory to the operating system.

    }
    catch ( IOException ioe )
    {

      System.out.println ( "ERROR: Server.disconnect (ServerSocket)" );
      System.out.println ( "       " + ioe );

    }

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns an InetAddress based on a 32-bit byte array while checking
      for exceptions. */

  public InetAddress getIPA ( byte [] address )
  {

    InetAddress ipa = null;

    try
    { // testable temporary IPA.

      ipa = InetAddress.getByAddress ( address );

    }
    catch ( UnknownHostException uhe )
    {

      System.out.println ( "ERROR: Server.getIPA (by byte array)" );
      System.out.println ( "       " + uhe );

    }

    return ipa;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns an InetAddress based on a String name while checking for
      exceptions. */

  public InetAddress getIPA ( String ipaStringName )
  {

    InetAddress ipa = null;

    try
    {

      ipa = InetAddress.getByName ( ipaStringName );

    }
    catch ( UnknownHostException uhe )
    {

      System.out.println ( "ERROR: Server.getIPA (by ipaStringName)" );
      System.out.println ( "       " + uhe );

    }

    return ipa;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns a Server Socket while checking for exceptions. */

  public ServerSocket getServerSocket ( int port, int backLog,
                                        InetAddress ipa )
  {

    ServerSocket socket = null;
    try
    {
      socket = new ServerSocket ( port, backLog, ipa );
    }
    catch ( IOException ioe )
    {
      System.out.println ( "ERROR 1: Server.getServerSocket" );
      System.out.println ( "         " + ioe );
    }
    catch ( SecurityException se )
    {
      System.out.println ( "ERROR 2: Server.getServerSocket" );
      System.out.println ( "         " + se );
    }
    catch ( IllegalArgumentException iae )
    {
      System.out.println ( "ERROR 3: Server.getServerSocket" );
      System.out.println ( "         " + iae );
    }
    catch ( NullPointerException npe )
    {
      System.out.println ( "ERROR 4: Server.getServerSocket" );
      System.out.println ( "         " + npe );
    }

    return socket;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns a Socket while checking for exceptions. */

  public Socket getSocket ()
  {

    Socket socket = null;
    try
    {

      socket = new Socket ();                               // create a socket.

    }
    catch ( IllegalArgumentException iae )
    {

      System.out.println ( "ERROR_1: Server.getSocket (Socket type 1)" );
      System.out.println ( "        " + iae );

    }
    catch ( SecurityException se )
    {

      System.out.println ( "ERROR_2: Server.getSocket (Socket type 1)" );
      System.out.println ( "        " + se );

    }

    return socket;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns a Socket while checking for exceptions. */

  public Socket getSocket ( String host, int port )
  {

    Socket socket = null;
    try
    {

      socket = new Socket ( host, port );                   // create a socket.

    }
    catch ( UnknownHostException uhe )
    {

      System.out.println ( "ERROR_1: Server.getSocket (Socket type 2)" );
      System.out.println ( "        " + uhe );

    }
    catch ( IOException ioe )
    {

      System.out.println ( "ERROR_2: Server.getSocket (Socket type 2)" );
      System.out.println ( "        " + ioe );

    }

    return socket;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns a formatted string product depicting an IPA with full
      leading zero padding on a 12 character field with interspersed dots.
      For example: "000.111.222.123" - 4 terms each with 3 characters. */

  public String ipaStringFormatter ( String rawIPAData )
  {

    final int MAX_TERMS = 4;                  // the number of terms in an IPA.
    final int NR = 3;        // the number of characters in a 3 character term.
    final String PREPEND = "    ";     // padding to fill out the IPA list box.
    final String DELIMITER = ".";

    String product = "";                           // the final output product.
    String [] ipaArray = split ( rawIPAData, DELIMITER, MAX_TERMS ); // divide.

    StringBuffer buffer = new StringBuffer ();      // a mutable string buffer.
    int charCount = 0;      // the count of numeric characters in the IPA term.
    int start = 0;          // position in the buffer to start the replacement.
    int end = 0;              // position in the buffer to end the replacement.
    for ( int currentTerm = 0; currentTerm < MAX_TERMS; currentTerm++ )
    {

      buffer = new StringBuffer ( "000" );   // create new buffer to work with.
      charCount = ipaArray [ currentTerm ].length ();
      start = NR - charCount;
      end = NR;
      buffer.replace ( start, end, ipaArray [ currentTerm ].toString () );

      if ( currentTerm == 0 )
      {

        product = PREPEND + buffer.toString () + ".";      // format.

      }
      else if ( currentTerm < ( MAX_TERMS - 1 ) )
      {

        product = product + buffer.toString () + ".";      // format.

      }
      else
      {

        product = product + buffer.toString ();            // format.

      }

    }

    System.gc ();        // clear away any garbage made during the above steps.

    return product;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this sets the current agent count. */

  public void setAgentCount ( int count )
  {

    this.agentCount = count;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this returns a string array built around matches of the given
     parameters. */

  public String [] split ( String toSplit, String splitter, int maxTerms )
  {

    // for example "000.111.222.123" with a dot splitter and 4 maxTerms.
    String [] result = new String [ maxTerms ];            // set aside memory.

    ArrayList < String > arrayList = new ArrayList < String > ();
    int pntr = 0;
    String string = "";
    do
    {

      if ( !toSplit.substring ( pntr, ( pntr + 1 ) ).equals ( splitter ) )
      { // around the splitter, begin to concatentate a result string.

        string = string + toSplit.substring ( pntr, ( pntr + 1 ) );

      }
      else
      { // the splitter "broke" the concatenation, so record the result.

        arrayList.add ( string );
        string = "";

      }

    }
    while ( ++pntr < toSplit.length () );
    arrayList.add ( string );  // record the last piece.
    Object [] object = arrayList.toArray ();

    for ( int i = 0; i < object.length; i++ )
    {

      result [ i ] = object [ i ].toString ();

    }

    return result;

  } // End method.
  // --------------------------------------------------------------------------

  // --------------------------------------------------------------------------
  /** this method does nothing. */

  public void template ()
  {
  } // End method.
  // --------------------------------------------------------------------------

// ----------------------------------------------------------------------------
}
